#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module handles general network settings
"""

from unicorn.core.module import Module
from unicorn.core.resource import UniqueValue, UniqueValueContainer
from unicorn.network.interface import Interface
from unicorn.network.host import Host, HostError
from types import IntType, StringType, BooleanType, UnicodeType

from unicorn.system.info import config as system_config
import logging


class _Module(Module):
    needs = [
        'unicorn.system.info',
        'unicorn.network.interface',
        'unicorn.network.host',
    ]

    def init(self):
        self.signals_map = [
            (config_signals.hostname.pre_update, lock_handler),
        ]

    def load(self, config_data):
        super(_Module, self).load(config_data)

        defaults = [
            ('workgroup', ''),
            ('domain', ''),
            ('proxy_add', u''),
            ('proxy_port', 0),
            ('proxy_login', u''),
            ('proxy_password', u''),
            ('hostname_locked', False),
        ]
        for attribute, default_value in defaults:
            if getattr(config, attribute) is None:
                setattr(
                    config,
                    attribute,
                    type(default_value)(
                        config_data.get(attribute, default_value)
                    )
                )

        Module.set_config(config, config_data, {
            'hostname': system_config.network_name,
        })

        current_ip = get_current_ip()
        if current_ip is None:
            # We have no IP address so don't create an entry in hosts. Check
            # if an entry with our hostname still exists, if so, delete it
            # (it means that we previously had an IP but we lost it).
            host = Host.find(host_name=unicode(config.hostname)).one()
            if host:
                host.delete()
        else:
            # Check if an host with our hostname or our IP already exists, 
            # if not, create it.
            host = Host.find(host_name=unicode(config.hostname)).one() \
                   or Host.find(ipv4_address=unicode(current_ip)).one()
            if not host:
                host = Host()

            # Always rewrite host parameters and save the resource. This way, 
            # post-update signal is triggered and /etc/hosts is rewritten.
            host.host_name = unicode(config.hostname)
            host.ipv4_address = unicode(current_ip)
            host.domain = unicode(config.domain)
            host.save()

        # Always re-generate /etc/hostname. We want this file to be
        # re-generated at each Unicorn startup, not only when the hostname 
        # is modified.
        hostname_change_listener(None, config.hostname)

def get_current_ip():
    """
    Returns the current ip of the product (default interface)
    @return: The current dotted decimal representation of the current ip or
             None if our product doesn't currently have an IP.
    """
    interface = Interface.find(Interface.is_active==True,
                               Interface.ipv4_address!=u'0.0.0.0').any()
    if interface:
        return interface.ipv4_address
    else:
        return None

def lock_handler(signal, sender):
    if config.hostname_locked:
        raise HostError, 'Hostname has been locked'

def hostname_change_listener(key, value):
    global config
    from logging import debug
    debug("Hostname was changed: %s" % (value, ))

    hostname_file = open('/etc/hostname', 'w')
    try: hostname_file.write('%s\n' % (value, ))
    finally: hostname_file.close()

    current_ip = get_current_ip()
    # No host entry should have an IP set to None 
    if current_ip is not None:
        host = Host.find(ipv4_address=current_ip).one()
        if host:
            host.host_name = unicode(value)
            host.save()

    from unicorn.core.service import InitNGProcess
    service = InitNGProcess('hostname')
    if service: service.restart()

def proxy_change_listener(key, value):
    pass

def wins_server_change_listener(key, value):
    pass

def dns_change_listener(key, value):
    pass

def workgroup_change_listener(key, value):
    pass

def domain_change_listener(key, value):
    global config
    for host in Host.all():
        if host.static: continue
        host.domain_name = unicode(value)
        host.save()

module = _Module(globals())
config = UniqueValueContainer(module,
                              UniqueValue(module, 'hostname', StringType,
                                          hostname_change_listener),
                              UniqueValue(module, 'proxy_add', UnicodeType,
                                          proxy_change_listener),
                              UniqueValue(module, 'proxy_port', IntType,
                                          proxy_change_listener),
                              UniqueValue(module, 'proxy_login', UnicodeType,
                                          proxy_change_listener),
                              UniqueValue(module, 'proxy_password', UnicodeType,
                                          proxy_change_listener),
                              UniqueValue(module, 'wins_server', StringType,
                                          wins_server_change_listener),
                              UniqueValue(module, 'primary_dns', StringType,
                                          dns_change_listener),
                              UniqueValue(module, 'secondary_dns', StringType,
                                          dns_change_listener),
                              UniqueValue(module, 'workgroup', StringType,
                                          workgroup_change_listener),
                              UniqueValue(module, 'domain', StringType,
                                          domain_change_listener),
                              hostname_locked = BooleanType,
                             )
config_signals = UniqueValueContainer.get_signals(config)
