#!/usr/bin/python
# -*- coding: utf-8 -*-
#-------------------------------------------------------------------------------
import os, time, sys
from types import BooleanType, StringType
from logging import warning

from storm.expr import Or, Select, Not

from unicorn.core.module   import Module
from unicorn.core.resource import UniqueValueContainer, UniqueValue, SaveError
from unicorn.core.service  import InitNGProcess
from unicorn.core.database import transaction_flush, in_transaction

from unicorn.authentication.user  import User
from unicorn.authentication.group import Group
from unicorn.authentication.local.user  import LocalUser
from unicorn.network.interface    import Interface
from unicorn.sharing.share        import Share, ShareUserAccess, \
                                         ShareGroupAccess
from unicorn.network.general      import config as network_config, \
                                         config_signals as network_config_signals
from unicorn.sharing.printing     import Server as PrintServer
from unicorn.storage.bind         import Binding
from unicorn.sharing.share_services import ShareService

class _Module(Module):
    # Samba should be started after 'unicorn.sharing.printing' (cups) to avoid
    # restarting samba on startup if a printer is already connected
    needs = [
        'unicorn.authentication.user',
        'unicorn.authentication.group',
        'unicorn.authentication.local.user',
        'unicorn.network.interface',
        'unicorn.network.general',
        'unicorn.sharing.share',
        'unicorn.sharing.share_services',
        'unicorn.sharing.printing',
        'unicorn.storage.bind'
    ]

    def init(self):
        self.signals_map = [
            (ShareService.post_update, signal_handler), 
            (Share.post_insert, signal_handler),
            (Share.post_update, signal_handler),
            (Share.post_delete, signal_handler),
            (ShareUserAccess.post_insert, signal_handler),
            (ShareUserAccess.post_update, signal_handler),
            (ShareUserAccess.post_delete, signal_handler),
            (ShareGroupAccess.post_insert, signal_handler),
            (ShareGroupAccess.post_update, signal_handler),
            (ShareGroupAccess.post_delete, signal_handler),
            (User.post_insert, signal_handler),
            (User.post_update, signal_handler),
            (User.post_delete, signal_handler),
            (Group.post_update, signal_handler),
            (Group.post_delete, signal_handler),
            (PrintServer.post_update, signal_handler),
            (network_config_signals.hostname.post_update, signal_handler),
            (network_config_signals.workgroup.post_update, signal_handler),
            (Interface.post_update, signal_handler),
            (transaction_flush, signal_handler),
        ]

    def load(self, config_data):
        global config

        super(_Module, self).load(config_data)

        # Read default values from the config file
        config.config_file_path = config_data.get('config_file_path',
                                                  'smb.conf')
        config.password_file_path = config_data.get('password_file_path',
                                                    'smbpasswd')
        config.service_name = config_data.get('service_name','smb')
        if config.enabled is None:
            config.enabled_on_default_shares = config_data.get('enabled_on_default_shares',
                                                                False)
        if config.enabled is None:
            # Value is not in the database, read it from the config file
            config.enabled = config_data.get('enabled', False)
        else:
            # Signal is not triggered, start the service if needed
            toggle_service(start=config.enabled)
        
        #cold start, create entries for shares that are already present in db
        subselect = Select(ShareService.share_id, 
                                       where="service='%s'" %config.service_name, 
                                       distinct=True)
        orphan_shares = Share.find(Not(Share.id.is_in(subselect)))
        
        for sh in orphan_shares:
            try:
                a = ShareService(sh.id, 
                                 unicode(config.service_name), 
                                 status=None)
                a.save()
            except SaveError:
                warning('Could not enable service %s on share %s'
                    ' (already exists?)' % (unicode(config.service_name), 
                                            sh.id))
                            
    def shutdown(self):
        global config, service

        super(_Module, self).shutdown()

        # Shutdown samba if needed
        if config.enabled: service.stop()

def toggle_service_listener(key, value): toggle_service(value)
def toggle_default_shares(key, value): 
    global service
    build_configuration()
    if config.enabled == True:
        service.reload()

module = _Module(globals())
config = UniqueValueContainer(
    module,
    UniqueValue(module, 'enabled', BooleanType, toggle_service_listener),
    UniqueValue(module, 'enabled_on_default_shares', BooleanType, toggle_default_shares),
    config_file_path=StringType,
    service_name=StringType,
    password_file_path=StringType,
)
config_signals = UniqueValueContainer.get_signals(config)
_restart = False
_restart_nmbd = False
_reload = False

class SambaService(object):
    samba = InitNGProcess('samba')
    smbd = InitNGProcess('samba/smbd')
    nmbd = InitNGProcess('samba/nmbd')
    winbind = InitNGProcess('samba/winbind')

    def __init__(self):
        self.services = (self.smbd, self.nmbd, self.winbind)

    def start(self):
        for s in self.services:
            s.start()

    def stop(self):
        for s in self.services:
            s.stop()

    def restart(self):
        for s in self.services:
            s.restart()

    def reload(self):
        self.samba.reload()

service = SambaService()

def toggle_service(start):
    global service
    if start:
        build_configuration()
        service.start()
    else:
        service.stop()

def signal_handler(signal, sender):
    """
        Handles signals relative to Samba configuration
    """
    global service, _restart, _restart_nmbd, _reload, config

    if signal == Share.post_insert:
        try:
            a = ShareService(sender.id, unicode(config.service_name), status=None)
            a.save()
        except SaveError:
            warning('Could not enable service %s on share %s'
                    ' (already exists?)' % (unicode(config.service_name), sender.id), 
                     exc_info=sys.exc_info())
    elif signal == ShareService.post_update:
        if sender.service != unicode(config.service_name):
            return
            
    if config.enabled:
        if signal == PrintServer.post_update:
            _restart = True
        elif signal == network_config_signals.hostname.post_update:
            _restart_nmbd = True
        elif signal != transaction_flush:
            _reload = True

        if not in_transaction() and (_restart or _restart_nmbd or _reload):
            build_configuration()

            if _restart:
                service.restart()
                _restart = False

            if _reload:
                if not _restart:
                    service.reload()
                _reload = False

            if _restart_nmbd:
                if not _restart and not _reload:
                    service.nmbd.restart()
                _restart_nmbd = False

            _dirty = False

def build_configuration():
    build_smbpasswd()
    build_smb_conf()

def build_smbpasswd():
    """
        Builds the smbpasswd file containing lmhash and nthash passwords
        for every user
    """
    global config

    smbpasswd = ''
    #------------------------------------------------------------------
    # Parse the list of users and write the corresponding entry in the
    # smbpasswd file
    #----------------
    for user in LocalUser.all():
        #------------------------------------------------------------------
        # The format of the smbpasswd file is as follows:
        #
        # <login>:<uid>:<lmhash>:<nthash>:[<flags>]:LCT-<last-changed-time>:
        #
        # See: http://www.manpagez.com/man/5/smbpasswd/ for more details
        #------------------------------------------------------------------
        if user.password_lmhash and user.password_nthash:
            smbpasswd += "%s:%s:%s:%s:[U          ]:LCT-%X:\n" % \
                          (user.system_login,
                           user.system_id,
                           user.password_lmhash,
                           user.password_nthash,
                           int(time.time()), )
        else:
            smbpasswd += "%s:%s:%s:%s:[NU         ]:LCT-%s:\n" % \
                          (user.system_login,
                           user.system_id,
                           'NO PASSWORDXXXXXXXXXXXXXXXXXXXXX',
                           'NO PASSWORDXXXXXXXXXXXXXXXXXXXXX',
                           '00000000', )

    smbpasswd_file = open(config.password_file_path, 'w')
    try: smbpasswd_file.write(smbpasswd.encode('utf-8'))
    finally: smbpasswd_file.close()

    os.chmod(config.password_file_path, 0600)

_conf_plugins = []

def register_conf_plugin(callback):
    """
        Adds a new plugin allowing altering Samba's global configuration
    """
    global _conf_plugins
    if callback not in _conf_plugins:
        _conf_plugins.append(callback)

def unregister_conf_plugin(callback):
    """
        Removes a global configuration plugin
    """
    global _conf_plugins
    if callback in _conf_plugins:
        _conf_plugins.remove(callback)

def format_smb_conf(conf_dict):
    """
        Formats a string corresponding to the smb configuration contained in the
        given dict
        @param conf_dict: a dict containing the configuration to format, the
                          following structure is expected:
        conf_dict = {
            section1: {
                'key1': 'value1',
                'key2': 'value2',
            },
            section2: {
                'key1': 'value1',
                'key2': 'value2',
                ...
            }
            ...
        }
    """
    # Let's have fun with some python string formatting kung-fu!
    return '\n'.join(
        [
            # Section header
            '[%s]\n        %s' % (
                section,
                # 8 spaces indentation
                '\n        '.join(
                    # Section content: key = value syntax
                    ' = '.join([k,conf_dict[section][k]]) \
                               for k in sorted(conf_dict[section])
                )
            ) for section in conf_dict
        ]
    )

def build_smb_conf():
    """
        Builds the smb.conf file containing global Samba parameters, as
        well as file shares definitions
    """
    global config

    wins_server = ''
    for interface in Interface.all():
        wins_server = interface.wins

    # printers spool dir
    var = Binding.find(name=u'var').one()
    if var and var.is_mounted():
        spool_dir = var.makedirs('spool/samba', 0777)
    else:
        spool_dir = '/var/spool/samba'

    smb_dict = {
        'global': {
            'server string': network_config.hostname or '',
            'workgroup': network_config.workgroup or '',
            'security': 'user',
            'map to guest': 'bad user',
            'guest account': 'nobody',
            'log file': '/var/log/samba/log.%%m',
            'max log size': '50',
            'log level' : '0',
            'socket options': 'TCP_NODELAY IPTOS_LOWDELAY SO_SNDBUF=65536 SO_RCVBUF=65536',
            'dns proxy': 'no',
            'local master':'yes',
            'os level': '20',
            'veto files': '/lost+found/Network Trash Folder/TheVolumeSettingsFolder/Temporary Items/',
            'unix charset': 'UTF-8',
            'use sendfile': 'yes',
            'use mmap': 'yes',
            'wide links': 'no',
            'unix extensions': 'no',
            'use spnego': 'yes',
            'load printers': 'yes',
            'printing': 'cups',
            'printcap name': 'cups',
            'disable spoolss': 'no',
            'null passwords': 'yes',
            'passdb backend': 'smbpasswd',
        },
        'printers' : {
            'use client driver': 'yes',
            'path': spool_dir or '',
            'browsable': 'no',
            'cups options': 'raw',
            'printable': 'yes',
            'guest ok': 'yes',
        }
    }
    if wins_server and wins_server != u'0.0.0.0':
        smb_dict['global']['wins server'] = wins_server

    for plugin in _conf_plugins:
        plugin(smb_dict)

    smb_conf = '%s\n' % format_smb_conf(smb_dict)

    if config.enabled_on_default_shares == True:
        # get shares with default config (status=None)
        # and those with ftp explicitely enabled (status=True)
        shares_list = [ share_srv.share
                        for share_srv
                        in ShareService.find(ShareService.service==unicode(config.service_name), 
                                             Or(ShareService.status==True, 
                                                 ShareService.status==None))
                        if share_srv.share.status==True]
    else:
        # get shares with ftp explicitely enabled (status=True)
        shares_list = [ share_srv.share
                       for share_srv
                       in ShareService.find(service=unicode(config.service_name), 
                                            status=True)
                       if share_srv.share.status==True]
                           
    #-----------------------------------------------------------------
    # Parse all shares to build the corresponding Samba configuration
    #-----------------------------------------------------------------
    for share in shares_list:
        #------------------------------------------------------------------
        # Check if the share is public, which happens if a ShareUserAccess
        # exists, with a user_id of 0
        #-----------------------------
        guest_acl = ShareUserAccess.find(user_id=0,
                                         share_id=share.id).one()

        #-------------------------------------------------------------
        # Retrieve the corresponding ACLs
        # First, get the list of users with a full access, then extend
        # it with the list of full access groups
        #----------------------------------------
        write_list = ['"%s"' % u for u in share.list_users_with_access(u'full')]
        write_list.extend('@"%s"' % g for g in share.list_groups_with_access(u'full'))

        #----------------------------------------------------------
        # Then, get the list of users with a read-only access, and
        # extend it with the list of read-only groups
        #---------------------------------------------
        read_list = ['"%s"' % u for u in share.list_users_with_access(u'read-only')]
        read_list.extend('@"%s"' % g for g in share.list_groups_with_access(u'read-only'))

        #-------------------------------------------------------------
        # Do not write share header if nobody has rights on this share
        #-------------------------------------------------------------
        if len(write_list) == 0 \
            and len(read_list) == 0 \
            and not guest_acl:
            continue;
        else:
            #--------------------------------------------------------
            # Write the share header, which is common for all shares
            #--------------------------------------------------------
            smb_conf += '[%s]\n' \
                    '        path = %s\n' \
                    '        vfs object = netatalk\n' \
                    '        browseable = yes\n' \
                    '        create mode = 0777\n' \
                    '        directory mask = 0777\n' % (share.name,
                                                         share.path)
                                                         
        #------------------------------------------------------------------
        # If the share is public, check if it is read-only or read-write,
        # and set the corresponding configuration
        #----------------------------------------
        if guest_acl:
            if share.volume.is_read_only is True:
                guest_acl_access = 'yes'
            else:
                guest_acl_access = guest_acl.access == u'read-only' \
                                                        and 'yes' or 'no'
            smb_conf += '        guest ok = yes\n' \
                        '        read only = %s\n' % (guest_acl_access, )
        else:
            #----------------------------------------------------------
            # Write the valid user list, which is a merge of read_list
            # and write_list, to the configuration file
            #-------------------------------------------
            if write_list or read_list:
                smb_conf += '        valid users = %s\n' % (' '.join(write_list+read_list), )
                            
        #------------------------------------------------------
        # Write the full access list to the configuration file
        #------------------------------------------------------
        if write_list:
            if share.volume.is_read_only is True \
                or (guest_acl and guest_acl.access == u'read-only'):
                smb_conf +=  '        writeable = no\n'
            else:
                smb_conf +=  '        writeable = yes\n'
            smb_conf+= '        write list = %s\n' % (' '.join(write_list), )
        #----------------------------------------------------
        # Write the read-only list to the configuration file
        #----------------------------------------------------
        if read_list:
            smb_conf += '        read list = %s\n' % (' '.join(read_list), )

    #-----------------------------------
    # Write the actual file on the disk
    #-----------------------------------
    smb_conf_file = open(config.config_file_path, 'w')
    try: smb_conf_file.write(smb_conf.encode('utf-8'))
    finally: smb_conf_file.close()

    #-----------------------------------------------------
    # Set the permissions on the file (0644 = -rw-r--r--)
    #-----------------------------------------------------
    os.chmod(config.config_file_path, 0644)
