#!/usr/bin/python
# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------

"""
This module handles auto update procedure
"""

from unicorn.core.module import Module
from storm.properties import Int, Unicode, Bool
from unicorn.core.database import create_table
from unicorn.core.resource import BaseResource
from unicorn.system.capsule import Capsule, CapsuleException
from unicorn.core.service import switch_runlevel
from unicorn.system.time import get_local_datetime

from unicorn.core.resource import UniqueValueContainer, UniqueValue
from types import IntType, StringType, BooleanType

# import for schedule_autoupdate
import os
import signal as signal_api
from logging import debug, error, info, warning
from subprocess import Popen
from datetime import timedelta

CRON_AUTOUPDATE_FILENAME = "/etc/cron.d/autoupdate"
CRON_AUTOUPDATE_WELCOME_FILENAME = "/etc/cron.d/autoupdate_welcome"

class ScheduledUpdateError(Exception):
    """ Exception for AutoUpdate """
    pass

class UpdateMode:
    """ Enum of the autoupdate modes """
    DISABLE         = 0
    AUTO_CHECK      = 1
    AUTO_DOWNLOAD   = 2
    AUTO_INSTALL    = 3

class UpdateStatusCode:
    """ Enum of the autoupdate status codes """
    AU_IDLE                  = 0
    AU_INITIALIZING          = 1
    AU_INITIALIZING_OVER     = 2
    AU_CHECKING_UPDATE       = 3
    AU_CHECKING_UPDATE_OVER  = 4
    AU_DOWNLOADING           = 5
    AU_DOWNLOADING_OVER      = 6

class UpdateErorCode:
    """ Enum of the autoupdate error codes """

    # Error codes inherited from the auto-update library
    AU_SUCCESS                  = 0  # Success
    AU_INIT_ERROR               = 1  # Generic Initialization Error
    AU_INIT_INVALID_PROTOCOL    = 2  # Invalid protocol given for update server
    AU_INIT_INVALID_VERSION     = 3  # Current product version string is invalid
    AU_INIT_INVALID_EXTENSION   = 4  # Update file extension is invalid
    AU_INIT_INVALID_DL_PATH     = 5  # Download path is invalid
    AU_INIT_INVALID_ROOT_PATH   = 6  # Update server root path is invalid
    AU_INIT_PLUGIN_OPEN_ERROR   = 7  # Update retrieval plugin could not be opened
    AU_CHECK_ERROR              = 8  # Generic Update checking error
    AU_CHECK_NO_UPDATE          = 9  # There are no update files on the server
    AU_GET_ERROR                = 10 # Generic Update retrieving error
    AU_GET_INVALID_PATH         = 11 # Update file path is invalid
    AU_GET_CONNECT_ERROR        = 12 # Cannot connect to update server
    AU_GET_DL_ERROR             = 13 # Cannot get data from update server
    AU_GET_INVALID_XML          = 14 # Downloaded XML description file is invalid
    AU_GET_INVALID_UPDATE       = 15 # Downloaded update file is invalid
    AU_GET_PROXY_CONF_ERROR     = 16 # Error while setting proxy configuration
    AU_SET_INVALID_PROXY_CONF   = 17 # Invalid proxy configuration
    AU_GET_INFO_ERROR           = 18 # Cannot retrieve update info
    AU_INTERRUPTED_BY_SIGNAL    = 19 # Update checking/downloading has been interrupted by a signal
    AU_SET_INVALID_SERVER_AUTH  = 20 # Invalid update server authentication structure
    AU_GET_SERVER_AUTH_ERROR    = 21 # Error while setting update server authentication configuration
    AU_SET_ERROR                = 22 # Generic error while setting an option

    # Error codes specific to Unicorn
    AU_CAPUSLE_EXTRACT_ERROR    = 100 # Error while extracting update capsule
    AU_AGENT_START_ERROR        = 101 # AutoUpdate Agent could not start properly

class _Module(Module):
    """ AutoUpdate module """
    def init(self):
        """
            Initialization, takes care of creating the auto update database
        """

        self.signals_map = [
            (AutoUpdateAgent.pre_update, signal_handler),
            (AutoUpdateAgent.post_insert, signal_handler),
        ]

        self.empty_db = create_table(
            'auto_update_agent',
            [
                'id INTEGER PRIMARY KEY',
                'agent_pid INTEGER',
                'update_available BOOLEAN',
                'update_error_code INTEGER',
                'update_status INTEGER',
                'download_progress INTEGER',
                'update_size INTEGER',
                'update_file_path VARCHAR',
                'last_update_check VARCHAR',
                'xml_product_name VARCHAR',
                'xml_product_version VARCHAR',
                'xml_build_revision VARCHAR',
                'xml_release_date VARCHAR',
                'xml_misc_info VARCHAR',
            ]
        )

    def load(self, config_data):
        """
            Second phase of the init, initializes the config parameters and
            create or retreive auto update agent ressource
        """
        global config
        super(_Module, self).load(config_data)

        # Always reload system parameters from unicorn.conf
        Module.set_config_from_conf(config, config_data, {
            'forbid_init_capsule': True,
            'update_dl_path': '',
            'update_server_login': '',
            'update_server_password': '',
            'update_file_extension': '',
            'update_dl_max_rate': 0,
            'update_server_version_path': '',
        })

        # All other values must be checked if exists in the db
        Module.set_config(config, config_data, {
            'update_server_address': '',
            'update_server_branch_path': '',
        })

        # create or start an auto update agent
        auto_update_agent_count = AutoUpdateAgent.all().count()
        if auto_update_agent_count == 0:
            # no auto_update_agent already exists, create a new one
            my_auto_update_agent = AutoUpdateAgent()
            # and reset attributes to default
            my_auto_update_agent.reset_attributes()
        else:
            # reapply status to restore last state
            my_auto_update_agent = AutoUpdateAgent.all().config(limit=1).one()

        debug('RESOURCE ID $$$ %d $$$ ' % my_auto_update_agent.id)

        Module.set_config(config, config_data, {
            'schedule_minute': 0,
            'schedule_hour': 3,
            'schedule_period': 1,
        })

        # Enable schedule_auto_update service if needed.
        if config.update_mode is None:
            config.update_mode = config_data.get(
                                    'update_mode',
                                    UpdateMode.DISABLE
                                    )
        elif config.update_mode != UpdateMode.DISABLE:
            autoupdate_cron_enable()

    def shutdown(self):
        """
            Called upon module shutdown, takes care of the necessary cleanup
        """
        super(_Module, self).shutdown()

def autoupdate_welcome_cron():
    """ Add autoupdate cron entry 5 minutes after welcome XP"""

    current_date = get_local_datetime() + timedelta(minutes=5)
    cron_autoupdate_str = '%d %d %d %d * root /usr/bin/auto_update_agent --syslog -i %d --mode %d --daemon' % (
        current_date.minute,
        current_date.hour,
        current_date.day,
        current_date.month,
        get_auto_update_id(),
        config.update_mode,
    )

    # Write autoupdate cron configuration.
    with open(CRON_AUTOUPDATE_WELCOME_FILENAME, "w") as autoupdate_fd:
        autoupdate_fd.write(cron_autoupdate_str + '\n')
    debug("Enable autoupdate welcome: %s", cron_autoupdate_str)

def autoupdate_cron_enable():
    """ Add autoupdate cron entry """

    if not (0 <= config.schedule_minute < 60):
        raise ScheduledUpdateError, "minute out of range (must be 0..59)"
    if not (0 <= config.schedule_hour < 24):
        raise ScheduledUpdateError, "hour out of range (must be 0..23)"

    if config.schedule_period == 1:
        cron_autoupdate_str = '%d %d * * * root /usr/bin/auto_update_agent --syslog -i %d --mode %d --daemon' % (
            config.schedule_minute,
            config.schedule_hour,
            get_auto_update_id(),
            config.update_mode,
        )
    elif config.schedule_period == 30:
        month_day = get_local_datetime().day
        if month_day > 28:
            month_day = 28
        cron_autoupdate_str = '%d %d %d * * root /usr/bin/auto_update_agent --syslog -i %d --mode %d --daemon' % (
            config.schedule_minute,
            config.schedule_hour,
            month_day,
            get_auto_update_id(),
            config.update_mode,
        )
    else: # default value is weekly (7 days)
        week_day = get_local_datetime().weekday()
        # For python datetime, sunday=6, monday=0... whereas for cron, sunday=0, monday=1...
        if week_day == 6:
            week_day = 0
        else:
            week_day = (week_day + 1) % 6

        cron_autoupdate_str = '%d %d * * %d root /usr/bin/auto_update_agent --syslog -i %d --mode %d --daemon' % (
            config.schedule_minute,
            config.schedule_hour,
            week_day,
            get_auto_update_id(),
            config.update_mode,
        )

    # Write autoupdate cron configuration.
    with open(CRON_AUTOUPDATE_FILENAME, "w") as autoupdate_fd:
        autoupdate_fd.write(cron_autoupdate_str + '\n')
    debug("Enable autoupdate: %s", cron_autoupdate_str)


def autoupdate_cron_disable():
    """
        Remove autoupdate cron entry
    """
    debug("Disable Auto Update Scheduled")

    if os.path.exists(CRON_AUTOUPDATE_FILENAME):
        os.unlink(CRON_AUTOUPDATE_FILENAME)

def toggle_service_handler(key, value):
    """
        React to update mode changes
    """
    if (value == UpdateMode.DISABLE) or (value == None):
        autoupdate_cron_disable()
    else:
        autoupdate_cron_enable()

def set_cron_autoupdate_handler(key, value):
    """
        React to auto update schedule mode parameter changes
    """
    if config.update_mode > UpdateMode.DISABLE:
        autoupdate_cron_enable()

module = _Module(globals())
config = UniqueValueContainer(
    module,

    UniqueValue(module, 'schedule_minute', IntType, set_cron_autoupdate_handler),
    UniqueValue(module, 'schedule_hour'  , IntType, set_cron_autoupdate_handler),
    UniqueValue(module, 'schedule_period', IntType, set_cron_autoupdate_handler),

    # Update mode :
    # - 0 : disable
    # - 1 : auto_check
    # - 2 : auto_check + auto_download
    # - 3 : auto_check + auto_download + auto_install
    UniqueValue(module, 'update_mode'    , IntType, toggle_service_handler),

    # Input param
    update_server_address       = StringType,
    update_server_version_path  = StringType,
    update_server_branch_path   = StringType,
    update_server_login         = StringType,
    update_server_password      = StringType,
    update_file_extension       = StringType,
    update_dl_path              = StringType,
    update_dl_max_rate          = IntType,
    forbid_init_capsule         = BooleanType,
)

def get_auto_update_id():
    """ Get the id of the current ressource in use """
    return AutoUpdateAgent.all().config(limit=1).one().id

def get_auto_update_res():
    """ Get the current ressource in use """
    return AutoUpdateAgent.all().config(limit=1).one()

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Download Agent
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
class AutoUpdateAgent(BaseResource):
    """ Handles Auto Update agent """

    __storm_table__ = 'auto_update_agent'

    id                      = Int(primary=True)
    agent_pid               = Int()
    update_available        = Bool()
    update_status           = Int()
    update_error_code       = Int()
    update_size             = Int()
    download_progress       = Int()
    update_file_path        = Unicode()
    last_update_check       = Unicode()
    xml_product_name        = Unicode()
    xml_product_version     = Unicode()
    xml_build_revision      = Unicode()
    xml_release_date        = Unicode()
    xml_misc_info           = Unicode()

    #--------------------------------------------------------------------------
    # Object Methods
    #--------------------------------------------------------------------------
    def __unicode__(self):
        return "[AutoUpdateAgent]"

    #--------------------------------------------------------------------------
    # Class constructor
    #--------------------------------------------------------------------------
    def __init__(self,  commit=True):

        # initiate the database to allow instance auto-commit
        BaseResource.__init__(self)
        self.save()

    def reset_attributes(self):
        self.agent_pid = 0
        self.update_available = False
        self.update_status = UpdateStatusCode.AU_IDLE
        self.update_error_code = UpdateErorCode.AU_SUCCESS
        self.update_size = 0
        self.download_progress = 0
        self.update_file_path = u''
        self.last_update_check = u''
        self.xml_product_name = u''
        self.xml_product_version = u''
        self.xml_build_revision = u''
        self.xml_release_date = u''
        self.xml_misc_info = u''

    def manual_check_update(self):
        """ Launch the auto_update_agent to check if an update is available """
        info('Start MANUAL CHECK UPDATE')
        args = [u'/usr/bin/auto_update_agent', u'--syslog', u'-i', str(self.id),
                u'--mode', str(UpdateMode.AUTO_CHECK), u'--daemon']
        try:
            process = Popen(
                args,
                executable = "/usr/bin/auto_update_agent",
                close_fds=True
            )
            # Wait for completion of the process that forks to check if the
            # agent has started successfully (remember that the agent is
            # launched with the --daemon option, so here we are waiting for
            # completion of the parent process that launches the daemon, not
            # the daemon process itself)
            ret = process.wait()
            # Check if process failed or was stopped by a signal
            if (os.WIFSIGNALED(ret)) or (os.WEXITSTATUS(ret) != 0):
                self.update_error_code = UpdateErorCode.AU_AGENT_START_ERROR
                self.save()
        except (OSError, ValueError):
            warning("Can't start Auto Update Agent")

    def manual_get_update(self):
        """ Launch the auto_update_agent to download the most recent update """
        info('Start MANUAL GET UPDATE')
        args = [u'/usr/bin/auto_update_agent', u'--syslog', u'-i', str(self.id),
            u'--mode', str(UpdateMode.AUTO_DOWNLOAD), u'--daemon']
        try:
            process = Popen(
                args,
                executable = "/usr/bin/auto_update_agent",
                close_fds=True
            )
            # Wait for completion of the process that forks to check if the
            # agent has started successfully (remember that the agent is
            # launched with the --daemon option, so here we are waiting for
            # completion of the parent process that launches the daemon, not
            # the daemon process itself)
            ret = process.wait()
            # Check if process failed or was stopped by a signal
            if (os.WIFSIGNALED(ret)) or (os.WEXITSTATUS(ret) != 0):
                self.update_error_code = UpdateErorCode.AU_AGENT_START_ERROR
                self.save()
        except (OSError, ValueError):
            warning("Can't start Auto Update Agent")

    def install_update_capsule(self):
        """
            Extract and install the previously downloaded capsule and then
            reboot
        """

        info('Start MANUAL INSTALL and REBOOT')
        upgrade_capsule = Capsule(self.update_file_path)
        try:
            upgrade_capsule.extract(no_init = config.forbid_init_capsule)
            upgrade_capsule.install()
            self.reset_attributes()
            switch_runlevel('reboot')
        except (CapsuleException, IOError) as e:
            error("Capsule extraction/installation error: %s" % str(e))
            self.update_error_code = UpdateErorCode.AU_CAPUSLE_EXTRACT_ERROR
            self.save()

    def cancel_update(self):
        """
            Cancel update checking/downloading by sending a signal to the
            Auto Update agent
        """

        info('Cancelling update checking/downloading')
        # Check if the Auto Update agent is running
        if self.agent_pid != 0:
            try:
                os.kill(self.agent_pid, signal_api.SIGTERM)
                info("SIGTERM sent to Auto Update agent (pid: %d)" %
                    self.agent_pid)
            except OSError:
                warning(
                    "SIGTERM could not be sent to Auto Update agent (pid: %d)" %
                    self.agent_pid
                )
        else:
            info('No Auto Update agent currently running')

def signal_handler(signal, sender):

    """ Check update_status has changed """
    fields = sender.get_updated_fields()
    if fields.has_key('update_status'):

        debug('update_status %s ' % sender.update_status)
        debug('update_available %s ' % sender.update_available)
        debug('update_error_code %s ' % sender.update_error_code)
        debug('config.update_mode %s ' % config.update_mode)

        if config.update_mode == UpdateMode.AUTO_INSTALL:
            if sender.update_status == UpdateStatusCode.AU_DOWNLOADING_OVER:
                if sender.update_error_code == UpdateErorCode.AU_SUCCESS:
                    debug(
                        'Update to install file path : %s',
                        sender.update_file_path
                    )
                    upgrade_capsule = Capsule(sender.update_file_path)
                    try:
                        upgrade_capsule.extract()
                        upgrade_capsule.install()
                    except (CapsuleException, IOError):
                        sender.update_error_code = UpdateErorCode.AU_CAPUSLE_EXTRACT_ERROR

        # This part is only for debug
        if config.update_mode == UpdateMode.AUTO_DOWNLOAD:
            if sender.update_status == UpdateStatusCode.AU_DOWNLOADING_OVER:
                if sender.update_error_code == UpdateErorCode.AU_SUCCESS:
                    debug('**** Update available to install **** ')

        if config.update_mode == UpdateMode.AUTO_CHECK:
            if sender.update_status == UpdateStatusCode.AU_CHECKING_UPDATE_OVER:
                if sender.update_available:
                    debug('**** Update available to download **** ')

    if fields.has_key('download_progress'):
        debug('Download progress : %d%% ' % sender.download_progress)

AutoUpdateAgent.init_signals(module)
