#! /usr/bin/env python

"""capsule launcher"""

from __future__ import with_statement

import logging
import sys
import os
import re
import glob
import tarfile
import hashlib
import xml.dom.minidom

from unicorn.core.module import Module
from unicorn.system.info import config as info_config
from unicorn.storage.disk import config as disk_config


class CapsuleModule (Module) :
    """integration of Capsule in Unicorn"""
    needs = ['unicorn.system.info', 'unicorn.storage.disk']

module = CapsuleModule (globals())


class TarFile (object) :
    """redefine TarFile to be with-compliant"""

    def __init__ (self, filename = None) :
        self.filename = filename

    def __enter__ (self) :
        self.tarfile = tarfile.open (self.filename)
        if not self.tarfile.getmembers () :
            raise CapsuleTarballException ('no tarball')
        return self.tarfile

    def __exit__ (self, exception_type, exception_value, traceback) :
        self.tarfile.close ()


class KernelPartition :
    """partitions for alternative kernel on each drive"""

    def __init__ (self) :
        self.partitions = []
        for drive_path in glob.iglob ('%s?%d' % (disk_config.internal_prefix, disk_config.rescue_kernel_partition)) :
            self.partitions.append (drive_path)

    def write (self, filename = None) :
        for partition_path in self.partitions :
            logging.debug ('write kernel %s to partition %s', filename, partition_path)
            with open (partition_path, 'w') as partition :
                with open (filename) as file :
                    partition.write (file.read())

    def erase (self) :
        for partition_path in self.partitions :
            logging.debug ('erase partition %s', partition_path)
            with open (partition_path, 'w') as partition :
                partition.write ('123456789')


class SystemVersion :
    """
    operations on system version number
    separators are all characters which are not digits
    separators are special if they contain "alpha", "beta" or "r" (not case sensitive)
    """

    def __init__ (self, version_string = None) :
        self.string = version_string
        self.parts = re.split ('([^0-9]+)', self.string)

    def __cmp__ (self, other) :
        result = 0
        separators = ('', '')
        is_separator = False
        for parts in map (None, self.parts, other.parts) :
            if is_separator :
                separators = parts
            else :
                result = self.__cmp_part (separators[0], parts[0], separators[1], parts[1])
                if result != 0 :
                    break
            is_separator = not is_separator
        return result

    def __cmp_part (self, self_sep, self_num, other_sep, other_num) :
        # alpha < beta < rc = r < nothing (major release) < .X (minor release)
        if not self_sep is None :
            self_sep = self_sep.lower ()
        if not other_sep is None :
            other_sep = other_sep.lower ()
        if self_sep != other_sep :
            # test special separators in order of release cycle
            result = self.__cmp_sep ('alpha', -1, self_sep, other_sep)
            if result != 0 :
                return result
            result = self.__cmp_sep ('beta', -1, self_sep, other_sep)
            if result != 0 :
                return result
            result = self.__cmp_sep ('r', -1, self_sep, other_sep)
            if result != 0 :
                return result
        # int(None) doesn't work, so replace None by -1
        if self_num is None:
            self_num = -1
        if other_num is None:
            other_num = -1
        return cmp(int(self_num), int(other_num))

    def __cmp_sep (self, pattern, self_result, self_sep, other_sep) :
        self_found = False
        other_found = False
        if self_sep is not None and self_sep.find (pattern) >= 0 :
            self_found = True
        if other_sep is not None and other_sep.find (pattern) >= 0 :
            other_found = True
        if (self_found and other_found) or (not self_found and not other_found) :
            return 0
        elif self_found :
            return self_result
        else :
            return self_result * -1


class CapsuleException (Exception) :
    pass
class CapsuleXMLException (CapsuleException) :
    pass
class CapsuleTarballException (CapsuleException) :
    pass
class CapsuleKernelException (CapsuleException) :
    pass
class CapsuleFileException (CapsuleException) :
    pass
class CapsuleHashException (CapsuleException) :
    pass
class CapsuleSystemVersionException (CapsuleException) :
    pass
class CapsuleProductException (CapsuleException) :
    pass


class CapsuleFile :
    """file extracted from a capsule"""

    def __init__ (self, filename = None) :
        self.filename = filename

    def check (self, original_hash = None) :
        """check file with SHA-1 hash"""
        logging.debug ('check SHA-1 hash of %s', self.filename)
        try :
            with open (self.filename) as file :
                computed_hash = hashlib.sha1 (file.read()).hexdigest()
        except Exception :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleFileException ('cannot compute hash of a file (%s)' % self.filename)
        if computed_hash != original_hash :
            logging.debug ('original hash = %s', original_hash)
            logging.debug ('computed hash = %s', computed_hash)
            raise CapsuleHashException ('file %s has a bad hash' % self.filename)


class CapsuleXML :
    """access to XML nodes from capsule"""

    LAST_TAG = '</Capsule>'

    def __init__ (self, filename = None) :
        self.filename = filename
        self.data = {}
        self.files = {}

    def parse (self) :
        logging.debug ('parse XML')
        document = xml.dom.minidom.parse (self.filename)
        self.rootNode = document.firstChild
        self.__get_files ()
        self.__get_basic_data ()
        self.__find_main_node ()
        self.__get_main_data ()
        return self.data

    def __get_files (self) :
        for file in self.rootNode.getElementsByTagName('File') :
            filename = file.firstChild.data
            hash = file.parentNode.getElementsByTagName('SHA1')[0].firstChild.data
            self.files[filename] = hash

    def __get_basic_data (self) :
        for node in self.rootNode.childNodes :
            if node.nodeName == 'Version' :
                self.version = node.firstChild.data
            elif node.nodeName == 'Main' :
                self.data['main_file'] = node.getElementsByTagName('File')[0].firstChild.data
            elif node.nodeName == 'Type' :
                self.data['type'] = node.firstChild.data
                self.type = '%s%s' % (self.data['type'][0].upper(), self.data['type'][1:])
            elif node.nodeName == 'Device' :
                for node in node.childNodes :
                    if node.nodeName == 'Platform' :
                        self.data['platform'] = node.firstChild.data
                    elif node.nodeName == 'Product' :
                        self.data['product'] = node.firstChild.data

    def __find_main_node (self) :
        for node in self.rootNode.childNodes :
            if node.nodeName == self.type :
                self.mainNode = node

    def __get_main_data (self) :
        for node in self.mainNode.childNodes :
            if node.nodeName == 'Version' :
                self.data['version'] = node.firstChild.data
            if node.nodeName == 'MinimalVersion' :
                self.data['min_version'] = node.firstChild.data


class Capsule :
    """
    capsule
       |__ XML
       \__ TAR
            |__ uImage
            |__ description.xml
            |__ description.sha1
            \__ repository/
                    |__ file1
                    |__ file2
                    |__ etc...
    """

    XML_FILENAME = 'description.xml'
    SHA1_FILENAME = 'description.sha1'
    DEFAULT_DIRECTORY = '/capsule'

    def __init__ (self, filename = None) :
        self.filename = filename
        self.extracted = False

    def extract (self, install_directory = None, no_init = False) :
        """extract files from capsule, check and return metadata"""
        logging.debug ('extract files from %s', self.filename)
        self.directory = install_directory or self.DEFAULT_DIRECTORY
        tarball_filename = self.__convert_to_tarball ()
        self.__extract_tarball (tarball_filename)
        self.__check_xml ()
        self.__parse_xml ()
        self.__check_files ()
        # If no_init is true, don't extract the capsule if it's an init capsule
        if no_init and (self.metadata.get ('type') == 'init'):
            raise CapsuleException ("Capsule not extracted: Extraction of init capsule is forbidden")
        self.__check_system_matching ()
        return self.metadata

    def __convert_to_tarball (self) :
        """convert capsule file to tarball"""
        with open (self.filename) as capsule :
            # remove XML header
            line = ''
            while not line.endswith ('%s\n' % CapsuleXML.LAST_TAG) :
                line = capsule.readline ()
                if not line :
                    raise CapsuleXMLException ('cannot find end of capsule header')
            # write caspule body as a tarball file
            if not os.access (self.directory, os.F_OK) :
                os.makedirs (self.directory)
            tarball_filename = os.path.join (self.directory, '%s.tar' % os.path.basename(self.filename))
            with open (tarball_filename, 'w') as tarball :
                tarball.write (capsule.read())
        # remove capsule
        os.remove (self.filename)
        return tarball_filename

    def __extract_tarball (self, tarball_filename) :
        """extract files from tarball"""
        try :
            with TarFile (tarball_filename) as tarball :
                tarball.extractall (self.directory)
        except Exception :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleTarballException ('cannot extract files from the tarball')
        self.extracted = True
        # remove tarball
        os.remove (tarball_filename)

    def __check_xml (self) :
        """check hash of XML file"""
        try :
            with open (os.path.join (self.directory, self.SHA1_FILENAME)) as hash_file :
                hash = hash_file.read (hashlib.sha1().digest_size * 2)
        except Exception :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleHashException ('cannot get hash (%s) of the XML file' % self.SHA1_FILENAME)
        try :
            CapsuleFile (os.path.join (self.directory, self.XML_FILENAME)).check (hash)
        except CapsuleFileException :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleXMLException ('cannot compute hash of the XML file (%s)' % self.XML_FILENAME)

    def __parse_xml (self) :
        """parse XML file"""
        try :
            self.xml = CapsuleXML (os.path.join (self.directory, self.XML_FILENAME))
            self.metadata = self.xml.parse ()
        except Exception :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleXMLException ('bad XML file (%s)' % self.XML_FILENAME)

    def __check_files (self) :
        """check hash of each file described in XML"""
        for filename, hash in self.xml.files.iteritems() :
            CapsuleFile (os.path.join (self.directory, filename)).check (hash)

    def __check_system_matching (self) :
        """check if capsule is intended to be run for this product/system"""
        logging.debug ('check if product/version is the right one')
        product = self.metadata.get ('product')
        if not product is None :
            if product != info_config.product_id :
                logging.debug ('capsule product (%s) is not %s', product, info_config.product_id)
                raise CapsuleProductException ('not built for %s' % info_config.product_id)
        version = self.metadata.get ('version')
        c_type = self.metadata.get ('type')
        if not version is None and c_type != 'init' :
            if SystemVersion(info_config.software_version) > SystemVersion(version) :
                logging.debug ('capsule version (%s) < %s', version, info_config.software_version)
                raise CapsuleSystemVersionException ('downgrade is forbidden')
        min_version = self.metadata.get ('min_version')
        if not min_version is None :
            if SystemVersion(info_config.software_version) < SystemVersion(min_version) :
                logging.debug ('capsule min_version (%s) > %s', min_version, info_config.software_version)
                raise CapsuleSystemVersionException ('system is too old to support this update')

    def install (self) :
        """install kernel to make the capsule effective at next reboot"""
        if not self.extracted :
            raise CapsuleException ('capsule not extracted')
        try :
            KernelPartition().write (os.path.join (self.directory, self.metadata['main_file']))
        except Exception :
            logging.debug ('Caught an exception', exc_info = sys.exc_info ())
            raise CapsuleKernelException ('cannot install kernel of capsule')
        logging.info ('capsule installed: %s   %s', self.filename, self.metadata)


if __name__ == '__main__' :
    # prepare caspule to be launched at next reboot
    logging.basicConfig (level = logging.DEBUG, format='%(levelname)-9s %(message)s')
    # handle command line arguments
    if len (sys.argv) < 2 :
        print 'syntax: %s <capsule file> [install directory]' % sys.argv[0]
        sys.exit (2)
    if len (sys.argv) > 2 :
        install_directory = sys.argv[2]
    else :
        install_directory = '/tmp/capsule'
    # decapsule
    capsule = Capsule (sys.argv[1])
    print capsule.extract (install_directory)
    capsule.install ()
